/********************************************************
 * cd_io_c -- Common object for all CD type I/O		*
 ********************************************************/
#include <sys/param.h>
#include <cstdio>
#include <unistd.h>
#include <glob.h>
#include <mntent.h>

#include "magic.h"
#include "backup.h"
#include "util.h"
#include "backup_info.h"
#include "tmp_file.h"
#include "log.h"
#include "proc.h"
#include "tty_lock.h"
#include "colon.h"
#include "cd_io_b.h"

/********************************************************
 * cd_io_c -- Create a new CD I/O object		*
 ********************************************************/
cd_io_c::cd_io_c(
    const std::string& name,	// The name of this IO
    const config& config	// Global configuration
)
{
    mount_dir = config.get_option(SECTION_IO, name, IO_CD_MOUNT_DIR);
    mount_dev = config.get_option(SECTION_IO, name, IO_CD_MOUNT_DEV);
    mount_cmd = config.get_option(SECTION_IO, name, IO_CD_MOUNT_CMD);
}

/********************************************************
 * mount -- Mount the cd.  Make sure it's done		*
 ********************************************************/
void cd_io_c::mount(
    const int cd_number	// Number of the cd to mount
)
{
    while (1) {
	if (is_mounted())
	    return;

	do_system(mount_cmd, LOG_IO);

	if (! is_mounted()) {
	    if (cd_number <= 0) 
		SYSTEM_ERROR("Could not mount CD");

	    tty_lock the_lock;	// Lock the CD

	    std::cout << "Mount cd " << cd_number << " and press <enter>:";

	    std::string line;
	    getline(std::cin, line);
	}
    }
}
/********************************************************
 * is_mounted -- Return true if the CD is mounted.	*
 ********************************************************/
bool cd_io_c::is_mounted(void)
{
    // The file containing the mount table
    FILE *mtab_file = setmntent("/etc/mtab", "r");

    while (1) {
	// The current mount entry
	struct mntent *cur_mount = getmntent(mtab_file);
	if (cur_mount == NULL)
	    break;
	
	// Check to see if the thing is mounted
	if (strcmp(cur_mount->mnt_fsname, mount_dev.c_str()) == 0) {
	    endmntent(mtab_file);
	    return (true);
	}
    }
    endmntent(mtab_file);
    return (false);
}
/********************************************************
 * umount -- Umount the device.  Make sure it gets 	*
 * total unmounted.					*
 ********************************************************/
void cd_io_c::umount(void)
{
    while (1) {
	// Already unmounted
	if (not is_mounted())
	    return;

	if (do_system("/bin/umount " + mount_dev, LOG_IO) == 0) 
	    return;

	{
	    tty_lock the_lock;	// Lock the console

	    std::cout << "Unable to umount " << mount_dev << std::endl;
	    std::cout << "Please umount and press <enter>:";

	    std::string answer;	// Answer from the input
	    std::getline(std::cin, answer);	// Wait for enter
	}
    }
}

/********************************************************
 * read_label -- Read a CD label.			*
 ********************************************************/
void cd_label::read_label(
    const std::string& label_dir	// Directory where the 
    					// cd is mounted
)
{
    // The file we are reading
    std::ifstream in_file((label_dir + "/label.txt").c_str());

    if (in_file.bad()) {
	SYSTEM_ERROR("Unable to open cd label file " + 
		label_dir + "/label.txt");
    }

    std::string line;	// Line from the input

    getline(in_file, line);
    if (line != CD_LABEL) 
	INTERNAL_ERROR("Expected CD label, got junk:" + line);

    serial = -1;
    date = "";
    cd_number = -1;
    last = false;

    colon_line parser(in_file);

    parser.expect_line(CD_LABEL_SERIAL, "Error reading CD label. Field SERIAL",
	    __FILE__, __LINE__);
    serial = parser.get_int_value("Error reading CD label. Field SERIAL");

    parser.expect_line(CD_LABEL_DATE, "Error reading CD label. Field DATE",
	    __FILE__, __LINE__);
    date = parser.get_value();

    parser.expect_line(CD_LABEL_CD, "Error reading CD label. Field CD",
	    __FILE__, __LINE__);
    cd_number = parser.get_int_value("Error reading CD label. Field CD");
    
    parser.expect_line(CD_LABEL_LAST, "Error reading CD label. Field LAST",
	    __FILE__, __LINE__);
    if (parser.get_int_value("Error reading CD label. Field LAST")) 
	last = true;
    else 
	last = false;
}
