/********************************************************
 * cd_io_r -- Handle the I/O for restoring from CD	*
 ********************************************************/
#include <algorithm>
#include <cstdio>
#include <mntent.h>

#include "magic.h"
#include "restore.h"
#include "log.h"
#include "proc.h"
#include "util.h"
#include "colon.h"
#include "tty_lock.h"
#include "cd_io_r.h"

/********************************************************
 * get_meta -- Read the meta data off cd and return	*
 * the file name.  Actually, since the CD is a file	*
 * system, just return the mounted file name.		*
 ********************************************************/
std::string cd_io_r::get_meta(void)
{
    if (is_mounted()) {
	umount();
    }
    mount(-1);

    label.read_label(mount_dir);
    return (mount_dir + "/meta.global");
}

/********************************************************
 * read -- Read the meta data from the file for the 	*
 * cd/io record.  The generic information has already	*
 * been read.						*
 ********************************************************/
void cd_io_info_r::read(
    std::istream& in_file		// The file to read
) 
{
    colon_line parser(in_file);	// Input parser/reader

    parser.expect_line(META_CD_FILE, "Error reading I/O type",
	    __FILE__, __LINE__);
    file_name = parser.get_value();

    std::string line;	// Line from the input

    getline(in_file, line);
    if (line != META_EOR)
	INTERNAL_ERROR("Expected EOR but got " + line);
}

/********************************************************
 * get_file -- Return the name of the file on the CD	*
 ********************************************************/
std::string cd_io_r::get_file(
    const io_info_r* const info		// The information about the batch
) {
    // The information about the CD
    const cd_io_info_r* const cd_info = 
	dynamic_cast<const cd_io_info_r* const>(info);

    load_cd(cd_info->media_sequence);

    // Name of the file
    std::string tmp_file = mount_dir + "/" + cd_info->file_name; 
    validate_file(tmp_file, info);
    return (tmp_file);
}

/********************************************************
 * eject -- Eject (or uneject a disk)			*
 ********************************************************/
void cd_io_r::eject(
    const IN_OUT flag	// Flag for eject direction
) {
    switch (flag) {
	case CD_EJECT_OUT:
	    do_system("eject " + mount_dev, LOG_CMD);
	    break;
	case CD_EJECT_IN:
	    do_system("eject -t " + mount_dev, LOG_CMD);
	    break;
	default:
	    INTERNAL_ERROR("Impossible flag");
    }
}
/********************************************************
 * load_cd -- Make sure that the given cd is loaded	*
 ********************************************************/
void cd_io_r::load_cd(
    const int cd_number	// Number of the cd to be loaded
) {
    while (cd_number != label.cd_number) {
	umount();
	eject(CD_EJECT_OUT);

	{
	    tty_lock the_lock;	// Lock console

	    std::cout << "Put cd #" << cd_number << 
		" in the drive and press return: ";
	    std::string response;
	    getline(std::cin, response);
	}

	eject(CD_EJECT_IN);
	label.read_label(mount_dir);
    }
}

/********************************************************
 * get_cur_media -- Return the currently loaded media	*
 * 	number.						*
 ********************************************************/
int cd_io_r::get_cur_media(void) {
    if (!is_mounted())
	mount(-1);
    label.read_label(mount_dir);
    return (label.cd_number);
}
