/********************************************************
 * collection_batch_r -- Module to handle the restore   *
 * 	part of a collection batch.			*
 ********************************************************/
#include <algorithm>
#include <iterator>
#include <sys/fcntl.h>
#include <unistd.h>

#include "util.h"
#include "magic.h"
#include "log.h"
#include "proc.h"
#include "colon.h"
#include "restore.h"
#include "collect_batch_r.h"

/********************************************************
 * collection_batch_info_r::collection_batch_info_r --	*
 * 	Create the batch information (and read the	*
 * 	initial values from the meta file.)		*
 ********************************************************/
collection_batch_info_r::collection_batch_info_r(
    const batch_info_r& info,		// Information about the batch
    std::istream& in_file		// The file to read the data from
): batch_info_r(info)
{
    colon_line parser(in_file);

    std::string line;	// Line from the input file

    getline(in_file, line);

    if (line != META_EOR) 
	INTERNAL_ERROR("Error: Collection batch information missing EOR");
}
/********************************************************
 * verify -- verify that the data on the backup is the	*
 * same as the data on the disk.			*
 ********************************************************/
void collection_batch_r::verify(
    const meta_r::info_set& gen_info,// Information about this batch
    io_r* const io 		     // The I/O to use for the verify
) {
    const collection_batch_info_r* const collection_info= 
	dynamic_cast<const collection_batch_info_r* const>
	((gen_info.begin())->batch_info);

    // The name of the file containing the batch
    std::string file_name = io->get_file(gen_info.begin()->io_info);

    std::string cmd; // Command to do the restore

    if (override_option) {
	// Command to do the verify 
	std::string verify_cmd = global_config.get_option(SECTION_BATCH, 
		collection_info->name, 
		BATCH_CMD_VERIFY_CMD);
	cmd = "cd " + backup_root + ";" + verify_cmd;
    } else {
	cmd = "cd " + backup_root + ";" + collection_info->verify_cmd;
    }

    log_msg(LOG_VERIFY, "Collect command: " + cmd);
    FILE *pipe_file = popen(cmd.c_str(), "w");
    if (pipe_file == NULL) 
	SYSTEM_ERROR("Could not create pipe for collection");

    // Open a file for this collection
    int in_fd = open(file_name.c_str(), O_RDONLY);
    if (in_fd <= 0)
	SYSTEM_ERROR("Could not open collection file");

    struct file_head head;	// head from the file

    if (read(in_fd, &head, sizeof(head)) != sizeof(head))
	SYSTEM_ERROR("Error reading file header");

    // Bytes left to read
    long unsigned int bytes_left = head.size;
    
    while (bytes_left > 0) {
	//TODO3: Make configurable the size below
	char buffer[62*1024];	// The buffer for the data

	// Size of the data to read
	unsigned long int to_read;

	if (bytes_left < sizeof(buffer)) 
	    to_read = bytes_left;
	else 
	    to_read = sizeof(buffer);

	// Size of the last read
	int read_size = read(in_fd, buffer, to_read);

	if (read_size <= 0)
	    break;

	if (std::fwrite(buffer, 1, read_size, pipe_file) != 
		static_cast<unsigned int>(read_size))
	    SYSTEM_ERROR("Could not write pipe");
	bytes_left -= read_size;
    }
    close(in_fd);

    fflush(pipe_file);
    pclose(pipe_file);
    pipe_file = NULL;

    io->rm_file(file_name);
}
/********************************************************
 * restore -- Restore a batch of files			*
 ********************************************************/
void collection_batch_r::restore(
     const meta_r::info_set& set, // Information on what to restore
     io_r* const io, 		// I/O to get the file from
     const string_set& files	// Files to restore
) {
    const collection_batch_info_r* const collection_info = 
	dynamic_cast<const collection_batch_info_r* const>
	((set.begin())->batch_info);

    // There should be only one item in a collection batch
    assert(set.size() == 1);

    // The name of the file containing the batch
    std::string file_name = io->get_file(set.begin()->io_info);

    std::string name_file; // The name of the file containing the files

    if (collection_info->names_needed) {
	// Files in this batch
	string_set files_this_batch;

	std::set_intersection(files.begin(), files.end(),
		collection_info->files.begin(), collection_info->files.end(),
		std::insert_iterator<string_set>(
			files_this_batch, files_this_batch.begin()));

	if(files_this_batch.size() == 0)
	    USER_ERROR("Attempt to restore by ID a batch that uses file names");

	// The name of the file containing the files
	name_file = write_name_file("rest.names", files_this_batch);
    }

    std::string cmd;	// Command to do restore

    if (override_option) {
	// Command to do the restore
	std::string restore_cmd = global_config.get_option(SECTION_BATCH, 
		collection_info->name, 
		BATCH_CMD_RESTORE_CMD);

	if (collection_info->names_needed) {
	    cmd = "name_file=" + name_file + ";" + restore_cmd; 
	} else {
	    cmd = restore_cmd;
	}
    } else {
	if (collection_info->names_needed) {
	    cmd = "name_file=" + name_file + ";" + collection_info->restore_cmd;
	} else {
	    cmd = collection_info->restore_cmd;
	}
    }
    log_msg(LOG_RESTORE, "Collect command: " + cmd);
    // File for the pipe to the restore command
    FILE *pipe_file = popen(cmd.c_str(), "w");
    if (pipe_file == NULL) 
	SYSTEM_ERROR("Could not create pipe for collection");

    // Open a file for this collection
    int in_fd = open(file_name.c_str(), O_RDONLY);
    if (in_fd <= 0)
	SYSTEM_ERROR("Could not open collection file");

    struct file_head head;	// head from the file

    if (read(in_fd, &head, sizeof(head)) != sizeof(head))
	SYSTEM_ERROR("Error reading file header");

    // Bytes left to read
    long unsigned int bytes_left = head.size;
    
    while (bytes_left > 0) {
	//TODO3: Make configurable the size below
	char buffer[62*1024];	// The buffer for the data

	// Size of the data to read
	unsigned long int to_read;

	if (bytes_left < sizeof(buffer)) 
	    to_read = bytes_left;
	else 
	    to_read = sizeof(buffer);

	// Size of the last read
	int read_size = read(in_fd, buffer, to_read);

	if (read_size <= 0)
	    break;

	if (std::fwrite(buffer, 1, read_size, pipe_file) != 
		static_cast<unsigned int>(read_size))
	    SYSTEM_ERROR("Could not write pipe");
	bytes_left -= read_size;
    }
    close(in_fd);

    fflush(pipe_file);
    pclose(pipe_file);
    pipe_file = NULL;

    io->rm_file(file_name);
    if ((!keep_option) && (collection_info->names_needed)) {
	if (unlink(name_file.c_str()) != 0)
	    SYSTEM_ERROR("Could not remove " + name_file);
    }
}
