/********************************************************
 * Functions to read the configuration file.		*
 ********************************************************/
#include <fstream>
#include <unistd.h>

#include "magic.h"
#include "validate.h"
#include "config.h"

// Things that can be whitespace
const char *const SPACES = " \t\n";

// Keyword table for verification
struct keyword_table_type {
    const char* const keyword;		  // Keyword indicating section type
    const struct valid_table* const table;// Validation information
} keyword_table[] = {
    {SECTION_OPTIONS,  	global_options_table},
    {SECTION_INPUT_LIST,input_list_table},
    {SECTION_BATCH, 	batch_table},
    {SECTION_IO, 	io_table},
    {SECTION_BACKUP, 	backup_table},
    {SECTION_JOB_LIST, 	job_table},
    {NULL, NULL}
};

/********************************************************
 * is_null -- Return true if this line is a null line	*
 * 	that contains nothing useful			*
 * 							*
 * (i.e. is a blank line or one with just a comment.)	*
 ********************************************************/
static bool is_null(
	const std::string line	// String to check for blankness
)
{
    // The position of the first non-blank character
    std::string::size_type cur_pos = 
	line.find_first_not_of(SPACES, 0);

    // Check for blank line 
    if (cur_pos == std::string::npos) {
	return (true);
    }
    // 	blank # comment
    if (line[cur_pos] == '#') {
	return (true);
    }
    return (false);
}
/********************************************************
 * single_config::read -- Read in a single config	*
 * section from a file.					*
 *							*
 * Reads in all the configuration data up to and	*
 * including the closing }.				*
 ********************************************************/
void single_config::read(
	std::istream &in_file,   // Stream to read
	const std::string i_type,// Name of the configuration
	const std::string i_name // Type of configuration
) 
{
    name = i_name;	// Store the name and type
    type = i_type;

    while (1) {
	std::string line;	// The line from the input

	std::getline(in_file, line);
	if (in_file.eof()) {
	    USER_ERROR("Unexpected eof reading data");
	}
	// At this point line is:
	// 1)	blank
	// 2) 	blank # comment
	// 3) 	keyword value
	// 4)	Closing curly	# (comment)
	//
	//
	// Blank (#1) or blank comment (#2) ?
	if (is_null(line)) {
	    continue;
	}

	std::string::size_type cur_pos = 0;	// Current position in the line
	cur_pos = line.find_first_not_of(SPACES, cur_pos);

	// 4)	Closing curly	# (comment)
	if (line[cur_pos] == '}') {
	    cur_pos = line.find_first_not_of(SPACES, cur_pos+1);
	    if (cur_pos == std::string::npos) {
		return;		// Closing curly and nothing
	    }
	    // Comment is the only thing that can follow
	    if (line[cur_pos] != '#') {
		USER_ERROR("Junk follows closing } on line");
	    }

	    return;	// Last line (ends with close curly )
	}

	// 3) Keyword value
	std::string::size_type last_pos;	// Last position in the line

	last_pos = line.find_first_of(SPACES, cur_pos);
	if (last_pos == std::string::npos) {
	    USER_ERROR("Could not find the end of the keyword");
	}
	// The keyword in the string
	std::string keyword = line.substr(cur_pos, last_pos-cur_pos);

	cur_pos = line.find_first_not_of(SPACES, last_pos);
	if (cur_pos == std::string::npos) {
	    USER_ERROR("Could not find value");
	}

	// Value can be "value" or
	// just value.  Let's see which
	std::string the_value;

	if (line[cur_pos] == '"') {
	    cur_pos++;
	    last_pos = cur_pos;
	    while (1) {
		// Termination character?
		if (line[last_pos] == '"') {
		    ++last_pos;
		    break;
		}

		// Check for the end of the value
		if (line[last_pos] == '\\') {
		    the_value += line[last_pos] + 1;
		    last_pos += 2;
		    continue;
		}
		the_value += line[last_pos];
		++last_pos;
	    }
	} else {
	    // Value is a simple keyword
	    last_pos = line.find_first_of(SPACES, cur_pos);
	    the_value = line.substr(cur_pos, last_pos-cur_pos);
	}
	last_pos = line.find_first_not_of(SPACES, last_pos);
	if (last_pos != std::string::npos) {
	    if (line[last_pos] != '#') {
		USER_ERROR("Junk at end of line");
	    }
	}
	data[keyword].push_back(the_value);
    }
}
/********************************************************
 * single_config::print -- Print the config		*
 * 	(for debugging and other diagnostics)		*
 ********************************************************/
void single_config::print(
    std::ostream &out_file	// The file to send the data to
) const
{
    out_file << "Type: " << type << " Name: " << name << std::endl;

    // Loop through each configuration item
    for(ConfigInfo::const_iterator cur_config = data.begin();
	    cur_config != data.end();
	    ++cur_config)
    {
	// Loop through each value
	ConfigInfo::value_type key_value = *cur_config;
	out_file << "    " << key_value.first << std::endl;

	// Loop through each item in the array
	for (string_vector::const_iterator cur_item = key_value.second.begin();
		cur_item != key_value.second.end();
		++cur_item)
	{
	    out_file << "        " << *cur_item << std::endl;
	}
    }
}

/********************************************************
 * config::read -- Read a configuration file		*
 ********************************************************/
void config::read(
   const char* const file_name
) {
    // This bit of paranoia is due to the fact that
    // the fstream class sometimes reports a file open
    // when it's not
    if (access(file_name, R_OK) != 0) 
	SYSTEM_ERROR(std::string("Could not access ") + file_name);

    // The input file
    std::ifstream in_file(file_name);

    if (in_file.bad()) {
	SYSTEM_ERROR(std::string("Could not open ") + file_name);
    }
    while (1) {
	std::string line;	// The line from the input

	getline(in_file, line);	// Get the input
	if (in_file.eof()) {
	    return;		// Done
	}

	if (is_null(line)) {
	    continue;
	}

	// Should be 	type name <curly>	# comment
	std::string::size_type cur_pos = 
	    line.find_first_not_of(SPACES, 0);

	if (cur_pos == std::string::npos) {
	    USER_ERROR("The line can't be blank at this point");
	}

	// End of the line
	std::string::size_type end_pos = 
	    line.find_first_of(SPACES, cur_pos);

	if (end_pos == std::string::npos) {
	    USER_ERROR("Expected spaces after type name");
	}

	// Found type name
	std::string type_name = line.substr(cur_pos, end_pos - cur_pos);

	cur_pos = line.find_first_not_of(SPACES, end_pos);
	if (cur_pos == std::string::npos) {
	    USER_ERROR("Missing name");
	}
	end_pos = line.find_first_of(SPACES, cur_pos);
	if (end_pos == std::string::npos) {
	    USER_ERROR("Expected space after name");
	}

	// The name of the configuration
	std::string config_name = line.substr(cur_pos, end_pos - cur_pos);

	cur_pos = line.find_first_not_of(SPACES, end_pos);
	if (cur_pos == std::string::npos) {
	    USER_ERROR("Missing open { ");
	}
	if (line[cur_pos] != '{') {
	    USER_ERROR("Expected {");
	}
	cur_pos = line.find_first_not_of(SPACES, cur_pos+1);
	if (cur_pos != std::string::npos) {
	    if (line[cur_pos] != '#') {
		USER_ERROR("Junk after config declaration");
	    }
	}

	config[type_name][config_name].read(in_file, type_name, config_name);

	// Loop through the keyword list, find this one, validate it
	keyword_table_type* cur_key;	// keyword pointer

	for (cur_key = keyword_table;
		cur_key->keyword != NULL;
		++cur_key) {

	    if (strcasecmp(cur_key->keyword, type_name.c_str()) == 0) {
		validate(type_name, config_name, cur_key->table);
		break;
	    }
	}
	if (cur_key->keyword == NULL)
	    USER_ERROR("Unknown section type " + type_name);
    }
}

/********************************************************
 * config::print -- print out a configuration		*
 * 	(for debugging)					*
 ********************************************************/
void config::print(
   std::ostream &out_file
) const 
{
    // Loop through the types and print them
    for (master_config_type::const_iterator cur_type = config.begin();
	cur_type != config.end();
	++cur_type) {
	
	// Loop through the names and print them
	for(config_set::const_iterator cur_name = cur_type->second.begin();
	    cur_name != cur_type->second.end();
	    ++cur_name) {
	        cur_name->second.print(out_file);
	}
    }
}

/********************************************************
 * get_option -- Given an option, return a value.	*
 * 							*
 * If the option does not exist, an exception is throw	*
 ********************************************************/
const std::string& config::get_option(
	const std::string& type,	   // Type name for lookup
	const std::string& name,	   // Name of the option for lookup
	const std::string& option,	   // Option name
	const unsigned int index /* = 0 */ // Index of the options
) const 
{
    // The pointer to the type part of the configuration
    master_config_type::const_iterator type_section = 
	config.find(type);
    
    // Test to see if the type was found
    if (type_section == config.end()) 
	USER_ERROR("Type " + type + "not found");

    // The section of the configuration we want
    config_set::const_iterator config_section = 
	type_section->second.find(name);

    if (config_section == type_section->second.end())
	USER_ERROR("Name " + name + " not found");

    return (config_section->second.get_option(option, index));
}

/********************************************************
 * get_option -- Get an option value.			*
 * Throws an exception if the option does not exit.	*
 ********************************************************/
const std::string& single_config::get_option(
	const std::string& keyword,		// Keyword of the option we want
	const unsigned int index /* = 0 */	// Which one we want
) const 
{
    // Iterator use to find the value
    ConfigInfo::const_iterator value_ptr;

    value_ptr = data.find(keyword);
    if (value_ptr == data.end()) {
	USER_ERROR("Keyword " + keyword + " not found for " + name + " " + type);
    }
    return (value_ptr->second[index]);
}
/********************************************************
 * get_option_all -- Get all option values.		*
 * Throws an exception if the option does not exit.	*
 ********************************************************/
const string_vector single_config::get_option_all(
	const std::string& keyword	// Keyword of the option we want
) const 
{
    // Iterator use to find the value
    ConfigInfo::const_iterator value_ptr;

    value_ptr = data.find(keyword);
    if (value_ptr == data.end()) {
	USER_ERROR("Keyword " + keyword + " not found");
    }
    return (value_ptr->second);
}

/********************************************************
 * defined -- Return true if the option is defined	*
 ********************************************************/
const bool config::defined(
	const std::string& type,		// Type name for lookup
	const std::string& name,		// Name of the option for lookup
	const std::string& option,		// Option name
	const unsigned int index /* = 0 */	// Index of the options
) const 
{
    // The pointer to the type part of the configuration
    master_config_type::const_iterator type_section = 
	config.find(type);
    
    // Test to see if the type was found
    if (type_section == config.end()) 
	return (false);

    // The section of the configuration we want
    config_set::const_iterator config_section = 
	type_section->second.find(name);

    if (config_section == type_section->second.end())
	return (false);

    return (config_section->second.defined(option, index));
}

/********************************************************
 * get_single -- Return a single configuration item	*
 ********************************************************/
const single_config& config::get_single(
	const std::string& type,	// Type name for lookup
	const std::string& name		// Name of the option for lookup
) const 
{
    // The pointer to the type part of the configuration
    master_config_type::const_iterator type_section = 
	config.find(type);
    
    // Test to see if the type was found
    if (type_section == config.end()) 
	USER_ERROR("Type " + type + " not found");

    // The section of the configuration we want
    config_set::const_iterator config_section = 
	type_section->second.find(name);

    if (config_section == type_section->second.end())
	USER_ERROR("Name " + name + " not found");

    return (config_section->second);
}

/********************************************************
 * validate -- Validate a single configuration		*
 * 							*
 * Throws exception on error.				*
 ********************************************************/
void config::validate(
    const std::string& type,	// Type name 
    const std::string& name,	// Option name
    const valid_table table[]	// Validation table
)
{
    // The configuration data
    single_config a_config = get_single(type, name);

    // Loop through each word in the keyword
    for (single_config::ConfigInfo::const_iterator 
	    cur_config =  a_config.data.begin();
	    cur_config != a_config.data.end();
	    ++cur_config) {
	// The keyword as a string
	const char* const keyword = cur_config->first.c_str();

	const valid_table* cur_table;
	/* Loop through the table and find the given keyword */
	for (cur_table = table; cur_table->keyword != NULL; ++cur_table) {

	    if (strcasecmp(keyword, cur_table->keyword) == 0) {

		if ((cur_table->flags & VALID_FLAG_MANY) == 0) {
		    if (cur_config->second.size() != 1)
			USER_ERROR("Configuration error: " + 
				type + " " + name +
				" Duplicate entry for " + 
				cur_config->first);
		}
		break;
	    }
	}
	if (cur_table->keyword == NULL) 
	    USER_ERROR("Configuration error: " +
		    type + " " + name +
		    " Illegal entry: " + cur_config->first);
    }
}
