/********************************************************
 * file_info -- Print information about a set of files	*
 * 	Debugging tool.					*
 * 							*
 * Usage: 						*
 * 	file_info <file> [<file> ...]			*
 ********************************************************/
#include <iostream>
#include <fstream>

#include "io_c.h"
#include "crc32.h"

/********************************************************
 * file_info -- print information about a single file	*
 ********************************************************/
void file_info(
    const char* const file_name	// File to get the info for
)
{
    std::cout << std::endl;
    std::cout << "File: " << file_name << std::endl;

    // The file to read
    std::ifstream in_file(file_name, std::ios::binary);

    if (in_file.bad()) {
	std::cerr << "Could not open " << file_name << std::endl;
	return;
    }

    struct file_head head;	// The head of the file

    in_file.read(reinterpret_cast<char*>(&head), sizeof(head));
    if (in_file.gcount() != sizeof(head))  {
	std::cerr << "Could not read head of " << file_name << std::endl;
	return;
    } 
    if (head.magic != HEAD_MAGIC)  {
	std::cout << "Bad head magic number for " << file_name << std::endl;
	return;
    }

    std::cout << "Backup number: " << head.backup_number << std::endl;
    std::cout << "Sequence " << head.media_sequence << ":" << 
	head.file_sequence << std::endl;
    std::cout << "Size " << head.size << std::endl;

    // Bytes left to read
    long int bytes_to_go = head.size;
    // Crc we've computed so far
    unsigned long int crc = 0;

    while (bytes_to_go > 0) {
	char buffer[60*1024];	// Buffer for the data
	unsigned long int read_size = bytes_to_go;	// Size of the next read

	if (read_size > sizeof(buffer))
	    read_size = sizeof(buffer);

	in_file.read(buffer, read_size);
	if (in_file.gcount() != static_cast<long int>(read_size)) {
	    std::cerr <<"Short read for " << file_name << std::endl;
	    return;
	}

	crc = crc32(crc, buffer, read_size);
	bytes_to_go -= read_size;
    }

    struct file_tail tail;	// The head of the file

    in_file.read(reinterpret_cast<char*>(&tail), sizeof(tail));
    if (in_file.gcount() != sizeof(tail))  {
	std::cerr << "Could not read tail of " << file_name << std::endl;
	return;
    }

    std::cout << "CRC " << head.crc << 
	((crc == head.crc) ? " Good" : " BAD!") <<
	std::endl;

    if (tail.magic != TAIL_MAGIC)
	std::cout << "Bad tail magic number " << std::endl;
}


int main(int argc, char *argv[])
{
    while (argc > 1) {
	file_info(argv[1]);
	++argv;
	--argc;
    }
    return (0);
}

