/********************************************************
 * io_r -- Handle the I/O for restore.  Most of the work	*
 * 	is done in the device handlers.			*
 ********************************************************/
#include <cstring>

#include "magic.h"
#include "io_c.h"
#include "io_b.h"
#include "config.h"
#include "tape_io_r.h"
#include "util.h"
#include "colon.h"
#include "crc32.h"
#include "cd_io_r.h"

/********************************************************
 * new_IO -- Create a new I/O object.  Which one 	*
 * depends on the type name in the configuration	*
 ********************************************************/
io_r* const new_io_r(
    const std::string& name, 	// Name of the object
    const config& config	// Configuration information on the object
)
{
    // The type of the object
    std::string type_name = config.get_option(SECTION_IO, name, IO_TYPE);

    if (str_eq(type_name, IO_TYPE_TAPE)) {
	return (new tape_io_r(name, config));
    }
    if (str_eq(type_name, IO_TYPE_CD)) {
	return (new cd_io_r(name, config));
    }
    USER_ERROR("Invalid type name " + type_name + " for I/O module");
}

/********************************************************
 * read_io -- Read an I/O identification string		*
 * and return a pointer to the information block	*
 ********************************************************/
io_info_r* const read_io(std::istream& in_file)
{
    std::string keyword;	// Keyword from the input
    std::string value;		// Value from the input
    std::string line;		// Line from the input

    std::string type;		// Type of the I/O
    int media_seq;		// Media sequence number
    int file_seq;		// File sequence number

    getline(in_file, line);
    if (line != META_IO_INFO) INTERNAL_ERROR("Expected INFO record");

    colon_line parser(in_file);	// Input parser/reader

    parser.expect_line(META_IO_TYPE, "Error reading I/O type",
	    __FILE__, __LINE__);
    type = parser.get_value();

    parser.expect_line(META_IO_MEDIA_SEQ, "Expected media sequence #",
	    __FILE__, __LINE__);
    media_seq = parser.get_int_value("Expected media sequence #");

    parser.expect_line(META_IO_FILE_SEQ, "Expected file sequence #",
	    __FILE__, __LINE__);
    file_seq = parser.get_int_value("Expected file sequence #");

    if (type == IO_TYPE_CD) {
	cd_io_info_r* info = new cd_io_info_r(media_seq, file_seq);
	info->read(in_file);
	return (info);
    } else if (type == IO_TYPE_TAPE) {
	tape_io_info_r* info = new tape_io_info_r(media_seq, file_seq);
	info->read(in_file);
	return (info);
    } else {
	INTERNAL_ERROR("Unknown media type in meta record");
    }
}

/********************************************************
 * validate_file -- Validate the head / tail of a file	*
 * 							*
 * Exception thrown if error.				*
 ********************************************************/
void validate_file(
    const std::string &tmp_file,	// File to validate
    const io_info_c* const io_info	// Information on the file
)
{
    // Serial number of the current backup
    static long unsigned int backup_number;	

    std::ifstream in_file(tmp_file.c_str(), std::ios::binary);

    if (in_file.bad())
	SYSTEM_ERROR("Could not open " + tmp_file);

    struct file_head head;	// The head of the file

    in_file.read(reinterpret_cast<char*>(&head), sizeof(head));
    if (in_file.gcount() != sizeof(head)) 
	SYSTEM_ERROR("Could not read head of " + tmp_file);

    if (head.magic != HEAD_MAGIC) 
	USER_ERROR("Bad magic number in file " + tmp_file);

     if (backup_number == 0) 
	 backup_number = head.backup_number;

     if (backup_number != head.backup_number)
	 USER_ERROR("Bad backup number in " + tmp_file);

     if (head.media_sequence != io_info->media_sequence)
	 USER_ERROR("Bad media sequence number " + tmp_file);

     if (head.file_sequence != io_info->file_sequence)
	 USER_ERROR("Bad file sequence number " + tmp_file);

     // Bytes left to read
     long int bytes_to_go = head.size;
     // Crc we've computed so far
     unsigned long int crc = 0;

     while (bytes_to_go > 0) {
	 char buffer[60*1024];	// Buffer for the data
	 unsigned long int read_size = bytes_to_go;	// Size of the next read

	 if (read_size > sizeof(buffer))
	     read_size = sizeof(buffer);

	 in_file.read(buffer, read_size);
	 if (in_file.gcount() != static_cast<long int>(read_size))
	     SYSTEM_ERROR("Short read for " + tmp_file);

	 crc = crc32(crc, buffer, read_size);
	 bytes_to_go -= read_size;
     }

    struct file_tail tail;	// The head of the file

    in_file.read(reinterpret_cast<char*>(&tail), sizeof(tail));
    if (in_file.gcount() != sizeof(tail)) 
	SYSTEM_ERROR("Could not read tail of " + tmp_file);

     if (crc != head.crc)
	 USER_ERROR("Bad file crc " + tmp_file);

     if (tail.magic != TAIL_MAGIC)
	 USER_ERROR("Bad tail magic for " + tmp_file);
}

