/********************************************************
 * restore -- Restore files				*
 ********************************************************/
#include <getopt.h>
#include <stdlib.h>
#include <fstream>
#include <sys/param.h>
#include <unistd.h>
#include <cstdio>
#include <algorithm>
#include <iterator>

#include "magic.h"
#include "tty_lock.h"
#include "util.h"
#include "config.h"
#include "tmp_file.h"
#include "filer.h"
#include "input_list.h"
#include "io_r.h"
#include "batch_r.h"
#include "log.h"
#include "exit.h"
#include "meta_r.h"
#include "restore.h"

config global_config;		// The configuration
meta_r global_meta;		// The global configuration information

bool keep_option = false;	// Keep temp files
bool do_log_option = false;	// If set, log things

std::ofstream log_file;		// File for logging things
std::string backup_root("/");	// Root of the backups

bool override_option = false;	// Override commands


// The name of the report file if any
static const char* report_file_name = NULL;

// The name of the log file, if any
static const char* log_file_name = NULL;

// Detailed report file
static const char* detail_report_name = NULL;

bool tty_lock::locked = false;	// TTY is currently locked
/********************************************************
 * usage -- Tell the user what to do			*
 ********************************************************/
static void usage(void)
{
    std::cerr << "Usage is:\n";
    std::cerr << "	restore toc [-k] [-l <log>] [-r <report>] [-x <detail-file>] [-d<info>] [-D<info>] [-O] <IO>" << std::endl;
    std::cerr << "	restore verify [-k] [-l <log>] [-r <report>] [-d<info>] [-D<info>] [-O] <IO>" << std::endl;
    std::cerr << "	restore restore <IO> [-k] [-f<file>] [-z<file>] [-C<dir>] -x  [-d<info>] [-D<info>] [-O] [file] [file]..." << std::endl;
    exit (EXIT_FATAL);
}
/********************************************************
 * check_batch -- Function object to check to see	*
 * if this batch is the one we are looking for.		*
 * 							*
 * See STL documentation / function object for 		*
 * why we are doing things this way.			*
 ********************************************************/
class check_batch: 
    public std::binary_function<meta_r::gen_info, unsigned int, bool> 
{
    public:
	bool operator()(
	    meta_r::gen_info info,	// Meta information to check
	    unsigned int batch_number	// Batch number to look for
	) const {
	    return (info.batch_info->batch_number == batch_number);
	}
};

/********************************************************
 * check_media -- Function object to check to see	*
 * if this batch is on the current media		*
 * 							*
 * See STL documentation / function object for 		*
 * why we are doing things this way.			*
 ********************************************************/
class check_media: 
    public std::binary_function<meta_r::gen_info, unsigned int, bool> 
{
    public:
	bool operator()(
	    meta_r::gen_info info,	// Information to check
	    unsigned int media		// Media to look for
	) const {
	    return (info.io_info->media_sequence == media);
	}
};
/********************************************************
 * find_next_batch -- find the next batch 		*
 * 	that's in the give information.			*
 * 							*
 * Use the indicated media as a hint as to which media	*
 * to use.						*
 *							*
 * Returns an extracted set of backup batches for	*
 * all with the same batch number, or a null		*
 * batch (result.size() == 0) if nothing found		*
 ********************************************************/
meta_r::info_set find_next_batch(
    const unsigned int requested_media,	// Media number requested
    meta_r::info_set& info		// Information to search (updated)
) {
    meta_r::info_set result;		// Result of our search

    // Pointer to the current information on a batch
    meta_r::info_set::iterator cur_set = find_if(
	    info.begin(), info.end(),
	    bind2nd(check_media(), requested_media));

    if (cur_set == info.end()) {
	// Could not find the correct media, so return the lowest
	cur_set = info.begin();
    }
    // Move the first batch onto the restore list
    result.insert(*cur_set);
    info.erase(cur_set);

    // The batch number we are looking for
    unsigned int this_batch = 
	result.begin()->batch_info->batch_number;

    while (1) {
	cur_set = std::find_if(info.begin(), info.end(),
		bind2nd(check_batch(), this_batch));

	if (cur_set == info.end())
	    break;

	result.insert(*cur_set);
	info.erase(cur_set);
    }
    /*
     * Log what happened
     */
    log_msg(LOG_SET, "");
    log_msg(LOG_SET, "set============================================");
    // Loop through each item in the result
    for (meta_r::info_set::const_iterator cur_info = result.begin();
	 cur_info != result.end();
	 ++cur_info)
    {
	log_msg(LOG_SET, "Media: " + 
	    int_to_string(cur_info->io_info->media_sequence) +
	    " File " + 
	    int_to_string(cur_info->io_info->file_sequence) + 
	    " Batch_number " + 
	    int_to_string(cur_info->batch_info->batch_number));
    }
    log_msg(LOG_SET, "todo--------------------------------------------");
    // Loop through each item in the result
    for (meta_r::info_set::const_iterator cur_info = 
		info.begin();
	 cur_info != info.end();
	 ++cur_info)
    {
	log_msg(LOG_SET, "Media: " + 
	    int_to_string(cur_info->io_info->media_sequence) +
	    " File " + 
	    int_to_string(cur_info->io_info->file_sequence) + 
	    " Batch_number " + 
	    int_to_string(cur_info->batch_info->batch_number));
    }
    log_msg(LOG_SET, "");

    return (result);
}

typedef enum {CMD_VER, CMD_TOC} SUB_CMD;	// Sub-command for toc/ver
/********************************************************
 * do_cmd -- Perform a toc or verify command		*
 ********************************************************/
static void do_cmd(
    const std::string& name,	// Name of the I/O to use
    const SUB_CMD sub_cmd	// Sub-command to execute
) {
    // The IO to use for reading the stuff
    io_r* io = new_io_r(name, global_config);

    // The name of the temp file in which the IO puts
    // the meta data after reading
    std::string meta_file = io->get_meta();

    global_meta.restore(meta_file);	// Restore the meta data

    switch (sub_cmd) {
	case CMD_TOC:
	    if (global_meta.id != META_FULL) {
		std::cerr << "WARNING: Last backup media not loaded" << std::endl;
	    }
	    if (report_file_name != NULL) {
		// File for the report
		std::ofstream report_file(report_file_name);
		if (report_file.bad()) {
		    SYSTEM_ERROR("Could not open report file: " + 
			    std::string(report_file_name));
		}
		global_meta.toc(report_file);
		report_file.close();
	    } else {
		global_meta.toc(std::cout);
	    }
	    if (detail_report_name != NULL) {
		global_meta.report(detail_report_name);
	    }
	    break;
	case CMD_VER:
	    if (global_meta.id != META_FULL) {
		USER_ERROR("WARNING: Last backup media not loaded");
	    }
	    {
		// The list of batchers available
		std::map<std::string, batch_r*> batch_list;
		
		// The sets of batches to verify 
		meta_r::info_set verify_todo = global_meta.info;

		bool first = true;
		// Loop through the records and do the verify
		while (verify_todo.size() != 0) {

		    // The data we are verifying
		    meta_r::info_set verify_batch = 
			find_next_batch(
				(first) ? 1: io->get_cur_media(), 
				verify_todo);
		    first = true;

		    // The type of the batch we are verifying
		    const std::string batch_type = 
			verify_batch.begin()->batch_info->type; 

		    // Create the batch if needed
		    if (batch_list.find(batch_type) == batch_list.end()) {
			batch_list[batch_type] = new_batch_r(batch_type);
		    }

		    batch_list[verify_batch.begin()->batch_info->type]->
			verify(verify_batch, io);
		}
	    }
	    break;
	default:
	    INTERNAL_ERROR("Impossible sub-cmd");
	    break;
    }
}

/********************************************************
 * do_toc_ver -- Do a table of contents or verify	*
 ********************************************************/
static void do_toc_ver(
    int argc,			// # args on the command line
    char *argv[],		// Arguments from the command line
    const SUB_CMD sub_cmd	// Sub-command to execute
) {
    argc--;
    argv++;	// Consume a command line argument

    while (1) {
	int opt = getopt(argc, argv, "x:kl:r:d:D:O");
	if (opt == EOF)
	    break;

	switch (opt) {
	    case 'd':
		init_log(optarg);
		break;
	    case 'D':
		init_console_log(optarg);
		break;
	    case 'k':
		keep_option = true;
		break;
	    case 'l':
		log_file_name = optarg;
		do_log_option = true;
		break;
	    case 'r':
		report_file_name = optarg;
		break;
	    case 'O':
		override_option = true;
		break;
	    case 'x':
		detail_report_name = optarg;
		break;
	    default:
		std::cerr << "Illegal option " << opt << std::endl;
		usage();
		break;
	}
    }

    if ((argc - optind) != 2) {
	usage();
    }
    if (log_file_name != NULL) {
	log_file.open(log_file_name, std::ios::out);
	if (log_file.bad()) {
	    std::cerr << "Error: Could not open log file " << 
		log_file_name << std::endl;
	    exit (EXIT_FATAL);
	}
    }

    global_config.read(argv[optind]);
    tmp_name_init(global_config);
    do_cmd(argv[optind+1], sub_cmd);
}

/********************************************************
 * read_files -- Read a list of files from the 		*
 * 	given input file and add it to the file list	*
 ********************************************************/
static void read_files(
    string_set& file_list,	// List of files (updated)
    const char *const file_name,// Name of the file to read
    const char eol		// End of line character
) {
    std::ifstream in_file(file_name);	// File to read

    if (in_file.bad()) 
     	USER_ERROR("Could not open file " + std::string(file_name));

    while (1) {
	std::string line;	// Line from the input

	getline(in_file, line, eol);
	if (in_file.eof())
	    break;

	file_list.insert(line);
    }
}
/********************************************************
 * do_restore -- Restore a set of files			*
 ********************************************************/
static void do_restore(
    int argc,			// # args on the command line
    char *argv[]		// Arguments from the command line
) 
{
    char *zero_from_file = NULL;	// File name file (zero eol)
    char *name_from_file = NULL;	// File name file (\n eol)
    char *restore_dir = NULL;		// Directory in which to do restore
    bool fake_it = false;		// Fake restore
    std::set<int> id_list;		// List of ids to backup

    argc--;
    argv++;	// Consume a command line argument
    /*
     * -C<dir>   -- Chdir to <x> before restore
     * -D<chars> -- Set debug information for log console
     * -d<chars> -- Set debug information for log file
     * -I<id>    -- Restore by backup id
     * -k        -- Keep temporary files
     * -l<file>  -- Log file
     * -n<file>  -- Get the files from the named file
     * -O        -- Cmds in config file override the ones on the backup
     * -x        -- Don't really restore
     * -z<file>  -- Get the files from the named file (zero record separator)
     */

    while (1) {
	int opt = getopt(argc, argv, "C:D:d:I:kl:n:Oxz:");
	if (opt == EOF)
	    break;

	switch (opt) {
	    case 'C':
		restore_dir = optarg;
		break;
	    case 'D':
		init_console_log(optarg);
		break;
	    case 'd':
		init_log(optarg);
		break;
	    case 'I':
		id_list.insert(atoi(optarg));
		break;
	    case 'k':
		keep_option = true;
		break;
	    case 'l':
		log_file_name = optarg;
		do_log_option = true;
		break;
	    case 'n':
		name_from_file = optarg;
		break;
	    case 'O':
		override_option = true;
		break;
	    case 'x':
		fake_it = true;
		break;
	    case 'z':
		zero_from_file = optarg;
		break;
	    default:
		std::cerr << "Illegal option " << 
		    static_cast<char>(opt) << std::endl;
		usage();
		break;
	}
    }

    if ((argc - optind) <= 1) {
	usage();
    }
    if (log_file_name != NULL) {
	log_file.open(log_file_name, std::ios::out);
	if (log_file.bad()) {
	    std::cerr << "Error: Could not open log file " << 
		log_file_name << std::endl;
	    exit (EXIT_FATAL);
	}
    }
    global_config.read(argv[optind]);
    ++optind;

    if (restore_dir != NULL) {
	if (chdir(restore_dir) != 0)
	    SYSTEM_ERROR(std::string("Could not chdir to ") + restore_dir);
    }

    tmp_name_init(global_config);

    // The IO to use for reading the stuff
    io_r* const io = new_io_r(argv[optind], global_config);
    ++optind;

    // The name of the temp file in which the IO puts
    // the meta data after reading
    std::string meta_file = io->get_meta();

    global_meta.restore(meta_file);	// Restore the meta data
    if (global_meta.id != META_FULL) {
	USER_ERROR("WARNING: Last backup media not loaded");
    }

    string_set files;	// The list of files to restore
    if (name_from_file != NULL) {
	read_files(files, name_from_file, '\n');
    }
    if (zero_from_file != NULL) {
	read_files(files, zero_from_file, '\0');
    }

    while (optind < argc) {
	files.insert(argv[optind]);
	optind++;
    }

    if ((files.size() == 0) && (id_list.size() == 0)) {
	files = global_meta.all_files();
    }  else {
	files = global_meta.match(files);
    }

    // The list of batchers available
    std::map<std::string, batch_r*> batch_list;

    // Set of batches to restore
    meta_r::info_set restore_todo = global_meta.get_batch_set(files);
    global_meta.add_ids(id_list, restore_todo);

    while (restore_todo.size() != 0) {

	// The batch to restore
	meta_r::info_set restore_batch = 
	    find_next_batch(io->get_cur_media(), restore_todo);

	// The type of the batch we are dealing with
	const std::string batch_type = 
	    restore_batch.begin()->batch_info->type; 

	// Create the batch if needed
	if (batch_list.find(batch_type) == batch_list.end()) {
	    batch_list[batch_type] = new_batch_r(batch_type);
	}

	batch_list[restore_batch.begin()->batch_info->type]->
	    restore(restore_batch, io, files);
    }
}
/*------------------------------------------------------*/
int main(int argc, char *argv[])
{
    try {
	if (argc <= 1) 
	    usage();

	if (strcasecmp(argv[1], "toc") == 0) {
	    do_toc_ver(argc, argv, CMD_TOC);
        } else if ((strcasecmp(argv[1], "ver") == 0) || 
	    (strcasecmp(argv[1], "verify") == 0)) {
	    do_toc_ver(argc, argv, CMD_VER);
        } else if ((strcasecmp(argv[1], "r") == 0) || 
	    (strcasecmp(argv[1], "restore") == 0)) {
	    do_restore(argc, argv);
	} else {
	    usage();
	}
    } 
    catch (user_error &err) {
	err.print();
    }
    catch (internal_error &err) {
	err.print();
    }
    catch (system_error &err) {
	err.print();
    }
    catch (generic_error &err) {
	std::cerr << "Internal error: Strange error caught" << std::endl;
	err.print();
    }
    return (EXIT_SUCCESS);
}
