/********************************************************
 * tape_file -- simulate a tape through the use		*
 * of multiple files.					*
 * 							*
 * This class is called when a tape name such as 	*
 * 	file:/tmp/test 					*
 * is used as a device name in a tape device 		*
 * specification.					*
 ********************************************************/
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <unistd.h>

#include "tape_file.h"
/********************************************************
 * tape_file::tape_file -- Create the tape file		*
 ********************************************************/
tape_file::tape_file(
    const std::string& device,	// Device name
    const int i_mode		// Mode
) : current_tape_number(1) {
    std::string keyword = device.substr(0, sizeof("File:")-1);

    if (strcasecmp(keyword.c_str(), "File:") != 0)
	INTERNAL_ERROR("Tape 'file' specification did begin with 'file:'");

    base_file_name = device.substr(sizeof("File:")-1);
    std::string::size_type colon_ptr = base_file_name.find_first_of(":");

    if (colon_ptr == std::string::npos) 
	USER_ERROR("File spec not in the form 'File:base:count'");

    max_file_number = atoi(base_file_name.substr(colon_ptr+1).c_str());
    base_file_name = base_file_name.substr(0, colon_ptr);

    struct stat file_info;	// Information on the file

    if (stat(base_file_name.c_str(), &file_info) != 0)
	SYSTEM_ERROR("Could not stat " + base_file_name);

    if (! S_ISDIR(file_info.st_mode)) 
	USER_ERROR(base_file_name + " is not a directory");

    switch (i_mode) {
	case O_RDWR:
	case O_WRONLY:
	    current_tape_number = 1;
	    current_file_number = 0;
	    break;
	case O_RDONLY:
	    current_tape_number = 1;
	    current_file_number = 0;
	    while (access(make_name().c_str(), F_OK) == 0) {
		++current_tape_number;
	    }
	    --current_tape_number;
	    break;
	default:
	    INTERNAL_ERROR("Unknown open mode in tape-file");
    }
}

/********************************************************
 * make_name -- Create a tape file name out of the	*
 * path, the tape number, and the file number.		*
 ********************************************************/
std::string tape_file::make_name(void) {
    char tape_str[5];
    sprintf(tape_str, "%02d", current_tape_number);

    char file_str[5];
    sprintf(file_str, "%03d", current_file_number);

    return (base_file_name + "/sim_tape." +
	    tape_str + "." + file_str);
}

/********************************************************
 * copy_file -- Copy a simulate tape file to disk or	*
 * 	disk to simulated tape.				*
 *********************************************************/
void tape_file::copy_file(
   const std::string &from,		// File to copy from
   const std::string &to,		// File to copy to
   const unsigned int blocking		// Blocking factor to use
)
{
    // Buffer for transfer
    char* buffer = new char[blocking];
    log_msg('T', "Copy file " + from + " => " + to);

    // Fd for input
    int in_fd = open(from.c_str(), O_RDONLY);
    if (in_fd < 0)
	SYSTEM_ERROR("Could not open " + from);

    // FD to write to
    int out_fd = open(to.c_str(), O_WRONLY|O_CREAT, 0777);
    if (out_fd < 0)
	SYSTEM_ERROR("Could not open " + to);

    while (1) {
	memset(buffer, '\0', blocking);

	// Size of the last read
	int read_size = read(in_fd, buffer, blocking);

	if (read_size <= 0) 
	    break;

	// Size of the write
	int write_size = write(out_fd, buffer, blocking);

	if (write_size != static_cast<int>(blocking))
	    SYSTEM_ERROR("Unable to write " + to);
    }
    delete[] buffer;
    close(in_fd);
    close(out_fd);
}

/********************************************************
 * copy_to_mt -- Copy a file to tape.			*
 ********************************************************/
int tape_file::copy_to_mt(
    const std::string& file_name,	// File to copy from
    const int block_size		// Blocking factor
)
{
    if (current_file_number > max_file_number) 
	return (-1);
    copy_file(file_name, make_name(), block_size);
    current_file_number++;
    return (0);
}
/********************************************************
 * copy_from_mt -- Copy file from tape to disk		*
 ********************************************************/
int tape_file::copy_from_mt(
    const std::string& file_name,	// File to put data in
    const int block_size		// blocking factor
)
{
    if (current_file_number > max_file_number) 
	return (-1);
    copy_file(make_name(), file_name, block_size);
    current_file_number++;
    return (0);
}

/********************************************************
 * change_tape -- Change the tape which is currently	*
 * loaded.						*
 ********************************************************/
bool tape_file::change_tape(
    const unsigned int tape_number	// tape number to switch to
) {
    current_tape_number = tape_number;
    current_file_number = 0;
    return (true);
}
